#include "Flock.h"
#include <cmath>

namespace Flock
{
  Dog::Dog() : Entity(),
	       VELOCITY(150.0f)
  {
    sprite = new SpriteAnimation("dog.png", FILTER_NONE, 64, 64);
    sprite->Add("idle", 4, 4, 0.35f);
    sprite->Add("walk", 0, 3, 5.0f);
    sprite->Play("idle");
    SetGraphic(sprite);

    AddTag("dog");
    
    SetCollider(new RectangleCollider(32, 32));
  }

  void Dog::Update()
  {
    Entity::Update();

    if (Input::IsKeyMaskHeld("left"))
      {
        scale.x = -1;
        velocity.x = -VELOCITY;
      }
    else if (Input::IsKeyMaskHeld("right"))
      {
        scale.x = 1;
        velocity.x = VELOCITY;
      }
    else
      velocity.x = 0;

    if (Input::IsKeyMaskHeld("up"))
      {
        velocity.y = -VELOCITY;
      }
    else if (Input::IsKeyMaskHeld("down"))
      {
        velocity.y = VELOCITY;
      }
    else
      velocity.y = 0;
    
    if (velocity == Vector2::zero)
      sprite->Play("idle");
    else
      sprite->Play("walk");


    position.x += velocity.x * Monocle::deltaTime;
    while (Collide("Solid"))
      position.x -= SIGN(velocity.x, 0.1);

    position.y += velocity.y * Monocle::deltaTime;
    while (Collide("Solid"))
      position.y -= SIGN(velocity.y, 0.1);


    if (position.y < GROUNDTOP - 5)
      position.y = GROUNDTOP - 5;

    if (position.y > GROUNDBOTTOM)
      position.y = GROUNDBOTTOM;

    FlockScene *s = (FlockScene *) scene;
    if (position.x < s->leftBoundary - 5)
      position.x = s->leftBoundary - 5;

    if (position.x > s->rightBoundary + 5)
      position.x = s->rightBoundary + 5;
    
    //Scene::GetCamera()->position.x = position.x;
  }

  Sheep::Sheep() : Entity(),
		   WALKSPEED(20.0f),
		   DOGSPEED(40.0f),
		   DOGDISTANCE(96),
		   WOLFSPEED(80.0f),
		   WOLFDISTANCE(64),
		   GENERALAWARENESS(128)
  {
    sprite = new SpriteAnimation("sheep1.png", FILTER_NONE, 64, 64);
    sprite->Add("idle", 0, 1, 0.35f);
    sprite->Add("walk", 2, 3, 5.0f);
    sprite->Add("trot", 2, 3, 10.0f);
    sprite->Add("freak", 4, 5, 10.0f);
    sprite->Add("chew", 6, 7, 5.0f);
    sprite->Play("idle");
    SetGraphic(sprite);

    baah = Assets::RequestAudio("baah.ogg");
    deck = Audio::NewDeck(baah);

    scaredogg = Assets::RequestAudio("scared.ogg");
    scared = Audio::NewDeck(scaredogg);

    AddTag("sheep");
    SetCollider(new RectangleCollider(32, 32));

    velocity = Vector2::Random() * WALKSPEED;

    //set aiTime to random in 2s so sheep move at different times
    aiTime= (float(rand()) / float(RAND_MAX)) * 2.0f;
  }

  void Sheep::Update()
  {
    Entity::Update();

    if (state == "dead")
      return;

    if (Collide("goal"))
      {
	//////////// play a nice noise//////////////
	RemoveSelf();
      }

    aiTime += Monocle::deltaTime;
    if (aiTime > 2.0f)
      {
	switch (rand() % 3)
	  {
	  case 1: // stand, blink
	    velocity = Vector2::zero;
	    sprite->Play("idle");
	    state = "idle";
	    break;
	  case 2: // move
	    velocity = Vector2::Random() * WALKSPEED;
	    sprite->Play("walk");
	    state = "wander";
	    
	    if (velocity.x < 0)
	      scale.x = -1;
	    else
	      scale.x = 1;
	    
	    break;
	  }

	std::list<Entity*> *tufts = scene->GetAllTag("tuft");
	if (tufts)
	  for (std::list<Entity*>::iterator i = tufts->begin(); i != tufts->end(); ++i)
	    {
	      Vector2 to_tuft = (*i)->position - position; //vector from sheep to wolf
	      if (fabs(to_tuft.GetMagnitude()) < GENERALAWARENESS)
		{
		  sprite->Play("walk");
		  state = "wander";
		  
		  velocity = to_tuft;
		  velocity.Normalize();
		  velocity *= WALKSPEED;

		  scale.x = SIGN(velocity.x, 1);
		}
	    }

	aiTime = 0.0f;
      }

    if (Collide("tuft"))
      {
	velocity = Vector2::zero;
	sprite->Play("chew");
      }

    Entity* dog = scene->GetFirstEntityWithTag("dog");
    Vector2 to_dog = dog->position - position; //vector from sheep to dog
    if (fabs(to_dog.GetMagnitude()) < DOGDISTANCE)
      {
	Vector2 run_away = -to_dog;
	run_away.Normalize();
	run_away *= DOGSPEED;
	velocity = run_away;

	if (state != "runningaway")
	  deck->Play();
	sprite->Play("trot");
	scale.x = SIGN(velocity.x, 1);

	state = "runningaway";
      }

    //run away from wolves
    std::list<Entity*> *wolves = scene->GetAllTag("wolf");
    if (wolves)
      for (std::list<Entity*>::iterator i = wolves->begin(); i != wolves->end(); ++i)
	{
	  Wolf *wolf = (Wolf *)(*i);
	  Vector2 to_wolf = wolf->position - position; //vector from sheep to wolf
	  if (fabs(to_wolf.GetMagnitude()) < WOLFDISTANCE)
	    {
	      sprite->Play("freak");
	      state = "freak";

	      scared->Play();
		
	      velocity = -to_wolf;
	      velocity.Normalize();
	      velocity *= WOLFSPEED;

	      scale.x = SIGN(velocity.x, 1);
	    }
	}

    position.x += velocity.x * Monocle::deltaTime;
    while ((velocity != Vector2::zero) && Collide("Solid"))
      {
	//std::cout << "x velocity: " << velocity.x << "\n";
	position.x -= SIGN(velocity.x, 0.1);
      }
    
    position.y += velocity.y * Monocle::deltaTime;
    while (Collide("Solid"))
      position.y -= SIGN(velocity.y, 0.1);


    if (position.y < GROUNDTOP)
      position.y = GROUNDTOP;

    if (position.y > GROUNDBOTTOM - 5)
      position.y = GROUNDBOTTOM - 5;

    FlockScene *s = (FlockScene *) scene;
    if (position.x < s->leftBoundary)
      position.x = s->leftBoundary;

    if (position.x > s->rightBoundary)
      position.x = s->rightBoundary;

    if ((position.x == s->rightBoundary || position.x == s->leftBoundary)
	&& (position.y == GROUNDBOTTOM - 5 || position.y == GROUNDTOP))
      position -= velocity.GetNormalized() * 64;
  }

  // WOLF

  Wolf::Wolf() : Entity(),
		 WALKSPEED(40.0f),
		 DOGSPEED(80.0f),
		 DOGDISTANCE(128),
		 SHEEPSPEED(80.0f),
		 SHEEPDISTANCE(192)
  {
    sprite = new SpriteAnimation("wolf.png", FILTER_NONE, 64, 64);
    sprite->Add("idle", 4, 4, 0.35f);
    sprite->Add("walk", 0, 3, 5.0f);
    sprite->Add("trot", 0, 3, 10.0f);
    sprite->Play("idle");
    SetGraphic(sprite);
    scale = Vector2(1.5, 1.5);

    AddTag("wolf");
    AddTag("removeOnTransition");
    SetCollider(new RectangleCollider(32, 32));

    velocity = Vector2::Random() * WALKSPEED;

    //set aiTime to random in 2s so sheep move at different times
    aiTime= (float(rand()) / float(RAND_MAX)) * 2.0f;
  }

  void Wolf::Update()
  {
    Entity::Update();

    aiTime += Monocle::deltaTime;
    if (aiTime > 2.0f)
      {
	switch (rand() % 3)
	  {
	  case 1: // stand, blink
	    velocity = Vector2::zero;
	    sprite->Play("idle");
	    state = "idle";
	    break;
	  case 2: // move
	    velocity = Vector2::Random() * WALKSPEED;
	    sprite->Play("walk");
	    state = "wander";
	    
	    if (velocity.x < 0)
	      scale.x = SIGN(-1, fabs(scale.x));
	    else
	      scale.x = SIGN(1, fabs(scale.x));
	    
	    break;
	  }
	aiTime = 0.0f;
      }

    Entity* dog = scene->GetFirstEntityWithTag("dog");
    Vector2 to_dog = dog->position - position; //vector from sheep to dog
    if (fabs(to_dog.GetMagnitude()) < DOGDISTANCE)
      {
	Vector2 run_away = -to_dog;
	run_away.Normalize();
	run_away *= DOGSPEED;
	velocity = run_away;

	/*if (state != "runningaway")
	  deck->Play();*/
	sprite->Play("trot");
	scale.x = SIGN(velocity.x, fabs(scale.x));

	state = "runningaway";
      }

    if (state != "pursuit" && state != "runningaway") {
      std::list<Entity*> *sheeps = scene->GetAllTag("sheep");
      for (std::list<Entity*>::iterator i = sheeps->begin(); i != sheeps->end(); ++i)
	{
	  Sheep *sheep = (Sheep *)(*i);
	  Vector2 to_sheep = sheep->position - position; //vector from wolf to sheep
	  if ((sheep->state != "dead") &&
	      fabs(to_sheep.GetMagnitude()) < SHEEPDISTANCE)
	    {
	      chasing = sheep;

	      /*if (state != "runningaway")
		deck->Play();*/
	      sprite->Play("trot");
	      
	      state = "pursuit";
	    }
	} // for each sheep
    } // if !chasing

    if (state == "pursuit")
      {	
	if (chasing->state == "dead")
	  {
	    state = "idle";
	    velocity = Vector2::zero;
	    chasing = NULL;
	    sprite->Play("idle");
	  }
	else
	  {
	    Vector2 to_sheep = chasing->position - position; //vector from wolf to sheep
	    
	    to_sheep.Normalize();
	    to_sheep *= SHEEPSPEED;
	    velocity = to_sheep;
	    
	    scale.x = SIGN(velocity.x, fabs(scale.x));
	  }

      }

    //don't walk on sheep
    if (Collide("sheep"))
      position -= velocity.GetNormalized() *= 0.1f;

    position.x += velocity.x * Monocle::deltaTime;
    while (Collide("Solid"))
      position.x -= SIGN(velocity.x, 0.1);

    position.y += velocity.y * Monocle::deltaTime;
    while (Collide("Solid"))
      position.y -= SIGN(velocity.y, 0.1);


    if (position.y < GROUNDTOP)
      position.y = GROUNDTOP;

    if (position.y > GROUNDBOTTOM)
      position.y = GROUNDBOTTOM;

    FlockScene *s = (FlockScene *) scene;
    if (position.x < s->leftBoundary + 10)
      position.x = s->leftBoundary + 10;

    if (position.x > s->rightBoundary - 10)
      position.x = s->rightBoundary - 10;
  }

  Background::Background() : Entity()
  {
    sprite = new Sprite("bg.png", FILTER_NONE, 800, 1024);
    SetGraphic(sprite);

    SetLayer(3);
  }

  void Background::Render()
  {
    Graphics::PushMatrix();
    Graphics::Translate(scene->GetCamera()->position);
    sprite->Render(this);
    Graphics::PopMatrix();
  }

  // Scene

  FlockScene::FlockScene() : Scene(),
			     transitioning(false),
			     win(false),
			     leftBoundary(10),
			     rightBoundary(790)
  {
  }
  
  void FlockScene::Begin()
  {
    Scene::Begin();

    Text *inst1 = new Text("Use arrow keys to move");
    inst1->position = Vector2(-380, -270);
    Add(inst1);

    Text *inst2 = new Text("Press ESC to quit");
    inst2->position = Vector2(-380, -250);
    Add(inst2);

    Input::DefineMaskKey("left", KEY_LEFT);
    Input::DefineMaskKey("right", KEY_RIGHT);
    Input::DefineMaskKey("up", KEY_UP);
    Input::DefineMaskKey("down", KEY_DOWN);

    // level editor
    Add( levelEditor = new LevelEditor() );
    levelEditor->Disable();

    // load level from files
    Level::LoadProject("project.xml");
    Level::Load("level.xml", this);

    Add(new Background());

    HideSpawners();

    Spawn(GetCamera()->position.x);

    std::list<Entity*> *tufts = GetAllTag("tuft");
    for (std::list<Entity*>::iterator i = tufts->begin(); i != tufts->end(); ++i)
      (*i)->SetCollider(new RectangleCollider(32, 32));

    std::list<Entity*> *rocks = GetAllTag("rock");
    for (std::list<Entity*>::iterator i = rocks->begin(); i != rocks->end(); ++i)
      (*i)->SetCollider(new RectangleCollider(64, 32));

    std::list<Entity*> *inv = GetAllTag("invisible");
    for (std::list<Entity*>::iterator i = inv->begin(); i != inv->end(); ++i)
      (*i)->SetCollider(new RectangleCollider(64, 64));
  }

  void FlockScene::Spawn(int x)
  {
    int left = x - 400;
    int right = x + 400;

    Entity* dogsp = GetFirstEntityWithTag("dogspawner");
    if (dogsp->position.x >= left and dogsp->position.x <= right)
      {
	Dog *dog = new Dog();
	dog->position = dogsp->position;;
	Add(dog);
      }

    std::list<Entity*> *spawners;
    spawners = GetAllTag("wolfspawner");
    for (std::list<Entity*>::iterator i = spawners->begin(); i != spawners->end(); ++i)
      if ((*i)->position.x >= left and (*i)->position.x <= right)
	{	
	  Wolf *wolf = new Wolf();
	  wolf->position = (*i)->position;
	  Add(wolf);
	}

    spawners = GetAllTag("sheepspawner");
    for (std::list<Entity*>::iterator i = spawners->begin(); i != spawners->end(); ++i)
      if ((*i)->position.x >= left and (*i)->position.x <= right)
	{
	  Sheep *sheep = new Sheep();
	  sheep->position = (*i)->position;
	  Add(sheep);
	}
  }

  void FlockScene::HideSpawners()
  {
    std::list<Entity*> *spawners = GetAllTag("spawner");
    for (std::list<Entity*>::iterator i = spawners->begin(); i != spawners->end(); ++i)
      {
	(*i)->isVisible = false;
      }
  }

  void FlockScene::ShowSpawners()
  {
    std::list<Entity*> *spawners = GetAllTag("spawner");
    for (std::list<Entity*>::iterator i = spawners->begin(); i != spawners->end(); ++i)
      {
	(*i)->isVisible = true;
      }
  }

  void FlockScene::Update()
  {
    Scene::Update();

    /*Vector2 ctr = Graphics::GetScreenCenter();
    Vector2 cam = GetCamera()->position;
    std::cout << "Center: x:" << ctr.x << " y:" << ctr.y;
    std::cout << " | Camera: x:" << cam.x << " y:" << cam.y << "\n";*/
    
    if (Input::IsKeyPressed(KEY_ESCAPE))
      Game::Quit();

    if (isPaused)
      {
	if (Input::IsKeyPressed(KEY_S) && Input::IsKeyHeld(KEY_LCTRL))
	  Level::Save();

	if (Input::IsKeyPressed(KEY_A))
	  Scene::GetCamera()->position.x -= 600;

	if (Input::IsKeyPressed(KEY_D))
	  Scene::GetCamera()->position.x += 600;
      }

    if (Input::IsKeyPressed(KEY_TAB))
      {
        // if we're not doing anything in the levelEditor...
        if (levelEditor->GetState() == FTES_NONE)
          {
            // toggle pause state
            isPaused = !isPaused;
            
            if (isPaused)
	      {
		ShowSpawners();
		levelEditor->Enable();
	      }
            else
	      {
		HideSpawners();
		levelEditor->Disable();
	      }
          }
      }

    // camera is @400 (middle)
    // let's put threshold at right quarter?
    int rightThreshold = GetCamera()->position.x + 200;
    
    // do transition if sheep are on right
    bool doTransition = true;
    std::list<Entity*> *sheeps = GetAllTag("sheep");
    for (std::list<Entity*>::iterator i = sheeps->begin(); i != sheeps->end(); ++i)
      {
	if ((*i)->position.x < rightThreshold)
	  doTransition = false;
      }    

    if (GetFirstEntityWithTag("dog")->position.x < rightThreshold)
      doTransition = false;

    if (doTransition && !isPaused && !transitioning)
      {
	//remove things which need to be removed before we recalc boundaries
	std::list<Entity*> *toRemove = GetAllTag("removeOnTransition");
	if (toRemove)
	  for (std::list<Entity*>::iterator i = toRemove->begin(); i != toRemove->end(); ++i)
	    {
	      if ((*i)->position.x <= rightBoundary)
		std::cout << "Removing an entity\n";
		(*i)->RemoveSelf();
	    }

	isPaused = true;
	transitioning = true;
	// new target is to move the screen but leave the quadrant on screen
	cameraTarget.x = GetCamera()->position.x + 600;
	leftBoundary = cameraTarget.x - 390;
	rightBoundary = cameraTarget.x + 390;

	Spawn(cameraTarget.x);

	Vector2 cam = GetCamera()->position;
	std::cout << "Moving camera from " << cam.x << ", " << cam.y;
	std::cout << " to target " << cameraTarget.x << ", " << cameraTarget.y << "\n";
      }

    if (transitioning)
      {
	//std::cout << "THE TRANSITION\n";
	if (GetCamera()->position.x == cameraTarget.x)
	  {
	    transitioning = false;
	    isPaused = false;
	  }
	else
	  {
	    /*Vector2 cam = GetCamera()->position;
	    std::cout << "Cam: x:" << cam.x << " y:" << cam.y << " | ";
	    std::cout << "Target: x:" << cameraTarget.x << " y:" << cameraTarget.y << "\n";*/
	    GetCamera()->position.x = APPROACH(GetCamera()->position.x, cameraTarget.x, 350 * Monocle::deltaTime);
	  }
      }

    std::list<Entity*> *sheep = GetAllTag("sheep");
    if (!sheep || sheep->empty())
      {
	isPaused = true;

	Text *win = new Text("You got the flock out!", 
			     Assets::RequestFont("LiberationSans-Regular.ttf", 50.0f));
	win->position = Vector2(-200,0);
	Add(win);
      }

  }// FlockScene::Update

  Text::Text(const std::string& text, FontAsset* font)
    : Entity(), text(text)
  {
    if (font == NULL)
      this->font = Assets::RequestFont("LiberationSans-Regular.ttf", 18.0f);
    else
      this->font = font;
  }

  void Text::Render()
  {
    Graphics::PushMatrix();

    // place text directly relative to camera
    Graphics::Translate(scene->GetCamera()->position + position);

    Graphics::SetBlend(BLEND_ALPHA);
    Graphics::SetColor(Color::white);
    Graphics::BindFont(font);
    
    Graphics::RenderText(*font, text, 0, 0);
    Graphics::PopMatrix();
  }

}
