#include "Monocle.h"
#include "Graphics/SpriteAnimation.h"
#include "LevelEditor/LevelEditor.h"

using namespace Monocle;

namespace Flock
{
  int const GROUNDTOP = 200;
  int const GROUNDBOTTOM = 590;
  // left & right are in scene...
  
  class Sheep : public Entity
  {
  public:
    Sheep();
    void Update();

    Vector2 velocity;
    float aiTime;
    std::string state;

    SpriteAnimation *sprite;
    AudioAsset *baah, *scaredogg;
    AudioDeck *deck, *scared;


    float const WALKSPEED;
    float const DOGSPEED; // speed when being herded
    int const DOGDISTANCE; // distance within which sheep runs from dog
    float const WOLFSPEED; // speed when running from wolf
    int const WOLFDISTANCE; // distance within which sheep runs from rolf
    int const GENERALAWARENESS; // distance for noticing interesting things
  };

  class Wolf : public Entity
  {
  public:
    Wolf();
    void Update();

    Vector2 velocity;
    float aiTime;
    std::string state;

    SpriteAnimation *sprite;
    Sheep *chasing;

    float const WALKSPEED;
    float const DOGSPEED; // speed when being herded
    int const DOGDISTANCE; // distance within which sheep runs from dog
    float const SHEEPSPEED; // speed when being chasing sheep
    int const SHEEPDISTANCE; // distance within which wolf chases sheep
  };

  class Dog : public Entity
  {
  public:
    Dog();
    void Update();

    Vector2 velocity;

    SpriteAnimation *sprite;

    float const VELOCITY;
  };

  class Background : public Entity 
  {
  public:
    Background();
    void Render();
    Sprite *sprite;
  };

  class FlockScene : public Scene
  {
  public:
    FlockScene();
    void Begin();
    void Update();
    void HideSpawners();
    void ShowSpawners();
    void Spawn(int x);

    LevelEditor *levelEditor;
    Vector2 cameraTarget;
    bool transitioning;
    bool win;

    int leftBoundary;
    int rightBoundary;
  };

  class Text: public Entity
  {
  public:
    Text(const std::string& text, FontAsset* font=NULL);
    
    void Render();
    
    void SetFont(FontAsset* font) { this->font = font; }
    void SetText(const std::string& text) { this->text = text; }
    
  protected:
    FontAsset* font;
    std::string text;
  };
}

